package com.android.launcher3.manager;


import android.app.Application;
import android.content.ContentProviderOperation;
import android.content.ContentResolver;
import android.content.ContentValues;
import android.content.Context;
import android.content.Intent;
import android.content.OperationApplicationException;
import android.database.Cursor;
import android.os.Handler;
import android.os.RemoteException;
import android.text.TextUtils;
import android.util.ArrayMap;

import com.android.launcher3.ItemInfo;
import com.android.launcher3.Launcher;
import com.android.launcher3.LauncherAppState;
import com.android.launcher3.LauncherProvider;
import com.android.launcher3.LauncherProviderChangeListener;
import com.android.launcher3.LauncherSettings;
import com.android.launcher3.SettingsActivity;
import com.android.launcher3.model.LoaderCursor;
import com.android.launcher3.touch.ItemClickHandler;
import com.android.launcher3.util.Logger;
import com.android.launcher3.util.Preconditions;

import java.util.ArrayList;
import java.util.List;

import io.reactivex.Observable;
import io.reactivex.Observer;
import io.reactivex.android.schedulers.AndroidSchedulers;
import io.reactivex.disposables.Disposable;
import io.reactivex.functions.Function;
import io.reactivex.schedulers.Schedulers;

import static com.android.launcher3.manager.Constants.DIALOG_KEY_VALUE;

/**
 * 桌面图标状态的管理工具
 */
public class LauncherManager {


    public static final String TAG = "LauncherManager";
    private ArrayMap<String, ManagementInfo> appInfo;

    private static volatile LauncherManager mInstance;

    private static Context mContext;

    private final LauncherProvider.ChangeListenerWrapper mListenerWrapper = new LauncherProvider.ChangeListenerWrapper();

    private Handler mListenerHandler;

    private static Class mSettingActivity;


    /**
     * 初始化方法 （必须!）
     *
     * @param application
     */
    public static void init(Application application) {
        init(application, Launcher.class);
    }

    /**
     * @param application
     * @param settingActivity 拦截系统设置后跳转到页面
     */
    public static void init(Application application, Class settingActivity) {
        mContext = application;
        mSettingActivity = settingActivity;
        LauncherManager.getInstance();

    }

    public static LauncherManager getInstance() {
        if (mInstance == null) {
            synchronized (LauncherManager.class) {
                if (mInstance == null) {
                    mInstance = new LauncherManager();
                }
            }
        }
        return mInstance;
    }

    private LauncherManager() {
        appInfo = new ArrayMap<>();
        mListenerHandler = new Handler(mListenerWrapper);
        loadManagementAppInfoFromDB();
    }

    /**
     * 获取受管理app的信息
     * @param packageName  应用包名
     * @return
     */
    public ManagementInfo getManageAppInfo(String packageName) {
        if (TextUtils.isEmpty(packageName))
            return null;
        if (appInfo.containsKey(packageName)) {
            return appInfo.get(packageName);
        }
        return null;
    }


    /**
     * Sets a provider listener.
     */
    public void setLauncherProviderChangeListener(LauncherProviderChangeListener listener) {
        Preconditions.assertUIThread();
        mListenerWrapper.mListener = listener;
    }

    /**
     * Overridden in tests
     */
    protected void notifyListeners() {
        mListenerHandler.sendEmptyMessage(LauncherProvider.ChangeListenerWrapper.MSG_LAUNCHER_PROVIDER_CHANGED);
    }

    /**
     * 增加某个app的信息
     */
    public void addPackageInfo(ManagementInfo managementInfo) {
        if (managementInfo == null)
            return;
        if (TextUtils.isEmpty(managementInfo.getPackageName()))
            return;
        if (!appInfo.containsKey(managementInfo.getPackageName())) {


            appInfo.put(managementInfo.getPackageName(), managementInfo);

            ContentValues contentValues = new ContentValues();

            long id = mContext.getContentResolver().call(ManagerAppProvider.DatabaseHelper.CONTENT_URI, ManagerAppProvider.METHOD_GENERATENEWITEMID, null, null).getLong(ManagerAppProvider.EXTRA_VALUE);
            contentValues.put(ManagerAppProvider.ManagementColumns._ID, id);
            contentValues.put(ManagerAppProvider.ManagementColumns.PACKAGENAME, managementInfo.getPackageName());
            contentValues.put(ManagerAppProvider.ManagementColumns.USE_TYPE, managementInfo.getUseType());
            contentValues.put(ManagerAppProvider.ManagementColumns.LOCK, managementInfo.getLock());


            ThreadpoolManager.getInstance().execute(new Runnable() {
                @Override
                public void run() {
                    mContext.getContentResolver().insert(ManagerAppProvider.DatabaseHelper.CONTENT_URI, contentValues);
                }
            });
            notifyListeners();
            Logger.e(TAG, "addPackageInfo");
        } else {
            updatePackageInfo(managementInfo);
        }


    }


    /**
     * 更新某个app的管理信息
     *
     * @param managementInfo
     */
    public void updatePackageInfo(ManagementInfo managementInfo) {

        appInfo.put(managementInfo.getPackageName(), managementInfo);

        ContentValues contentValues = new ContentValues();

        contentValues.put(ManagerAppProvider.ManagementColumns.PACKAGENAME, managementInfo.getPackageName());
        contentValues.put(ManagerAppProvider.ManagementColumns.USE_TYPE, managementInfo.getUseType());
        contentValues.put(ManagerAppProvider.ManagementColumns.LOCK, managementInfo.getLock());

        ThreadpoolManager.getInstance().execute(new Runnable() {
            @Override
            public void run() {
                mContext.getContentResolver().update(ManagerAppProvider.DatabaseHelper.CONTENT_URI, contentValues, ManagerAppProvider.ManagementColumns.PACKAGENAME + "='" + managementInfo.getPackageName() + "'", null);
            }
        });

        notifyListeners();
        Logger.e(TAG, "updatePackageInfo");
    }

    /**
     * 移除某个app的信息
     *
     * @param packageName
     */
    public void removePackageInfo(String packageName) {

        if (appInfo.containsKey(packageName)) {
            appInfo.remove(packageName);
            ThreadpoolManager.getInstance().execute(new Runnable() {
                @Override
                public void run() {
                    mContext.getContentResolver().delete(ManagerAppProvider.DatabaseHelper.CONTENT_URI, ManagerAppProvider.ManagementColumns.PACKAGENAME + "='" + packageName + "'", null);
                }
            });

            notifyListeners();
            Logger.e(TAG, "removePackageInfo");
        }
    }

    /**
     * 列表式添加
     *
     * @param appInfoList
     */
    public void addPackageInfoList(List<ManagementInfo> appInfoList) {
        if (appInfoList == null)
            return;

        ArrayList<ContentProviderOperation> ops = new ArrayList<ContentProviderOperation>();
        for (ManagementInfo managementInfo : appInfoList) {
            if (appInfo.containsKey(managementInfo.getPackageName())) {


                ContentValues contentValues = new ContentValues();

                contentValues.put(ManagerAppProvider.ManagementColumns.PACKAGENAME, managementInfo.getPackageName());
                contentValues.put(ManagerAppProvider.ManagementColumns.USE_TYPE, managementInfo.getUseType());
                contentValues.put(ManagerAppProvider.ManagementColumns.LOCK, managementInfo.getLock());

                ops.add(ContentProviderOperation.newUpdate(ManagerAppProvider.DatabaseHelper.CONTENT_URI).withValues(contentValues).withSelection(ManagerAppProvider.ManagementColumns.PACKAGENAME + "=?", new String[]{managementInfo.getPackageName()}).build());

            } else {

                ContentValues contentValues = new ContentValues();

                long id = mContext.getContentResolver().call(ManagerAppProvider.DatabaseHelper.CONTENT_URI, ManagerAppProvider.METHOD_GENERATENEWITEMID, null, null).getLong(ManagerAppProvider.EXTRA_VALUE);
                contentValues.put(ManagerAppProvider.ManagementColumns._ID, id);
                contentValues.put(ManagerAppProvider.ManagementColumns.PACKAGENAME, managementInfo.getPackageName());
                contentValues.put(ManagerAppProvider.ManagementColumns.USE_TYPE, managementInfo.getUseType());
                contentValues.put(ManagerAppProvider.ManagementColumns.LOCK, managementInfo.getLock());

                ops.add(ContentProviderOperation.newInsert(ManagerAppProvider.DatabaseHelper.CONTENT_URI).withValues(contentValues).build());
            }

            appInfo.put(managementInfo.getPackageName(), managementInfo);
        }

        notifyListeners();

        DBManager.doSql(mContext, ops);

    }


    /**
     * 批量删除
     *
     * @param dataList
     */
    public void deletePackageInfoList(List<ManagementInfo> dataList) {
        if (dataList == null)
            return;

        ArrayList<ContentProviderOperation> ops = new ArrayList<ContentProviderOperation>();
        for (ManagementInfo managementInfo : dataList) {
            if (appInfo.containsKey(managementInfo.getPackageName())) {
                appInfo.remove(managementInfo.getPackageName());
            }
            ops.add(ContentProviderOperation.newUpdate(ManagerAppProvider.DatabaseHelper.CONTENT_URI).withSelection(ManagerAppProvider.ManagementColumns.PACKAGENAME + "=?", new String[]{managementInfo.getPackageName()}).build());
        }
        notifyListeners();
        DBManager.doSql(mContext, ops);

    }


    private void loadManagementAppInfoFromDB() {
        Observable.just("").map(new Function<String, ArrayMap>() {
            @Override
            public ArrayMap apply(String s) throws Exception {


                ArrayMap<String, ManagementInfo> arrayMap = new ArrayMap<>();
                final ContentResolver contentResolver = mContext.getContentResolver();
                final Cursor c = contentResolver.query(
                        ManagerAppProvider.DatabaseHelper.CONTENT_URI, null, null, null, null);

                final int packageNameIndex = c.getColumnIndexOrThrow(
                        ManagerAppProvider.ManagementColumns.PACKAGENAME);
                final int useTypeIndex = c.getColumnIndexOrThrow(
                        ManagerAppProvider.ManagementColumns.USE_TYPE);
                final int lockIndex = c.getColumnIndexOrThrow
                        (ManagerAppProvider.ManagementColumns.LOCK);

                while (c.moveToNext()) {

                    String packageName = c.getString(packageNameIndex);
                    int useType = c.getInt(useTypeIndex);
                    int lock = c.getInt(lockIndex);

                    ManagementInfo managementInfo = new ManagementInfo();
                    managementInfo.setPackageName(packageName);
                    managementInfo.setUseType(useType);
                    managementInfo.setLock(lock);

                    arrayMap.put(packageName, managementInfo);

                }
                c.close();
                return arrayMap;
            }
        }).observeOn(AndroidSchedulers.mainThread()).subscribeOn(Schedulers.io()).subscribe(new Observer<ArrayMap>() {
            @Override
            public void onSubscribe(Disposable d) {

            }

            @Override
            public void onNext(ArrayMap arrayMap) {
                appInfo.putAll(arrayMap);
                Logger.e(TAG, "onNext");
            }

            @Override
            public void onError(Throwable e) {

                Logger.e(TAG, e.getLocalizedMessage());

            }

            @Override
            public void onComplete() {

            }
        });

    }

    /**
     * 获取受管理的app信息
     *
     * @return
     */
    public ArrayMap<String, ManagementInfo> getAppInfo() {
        return appInfo;
    }

    public void backLauncherHome(Context context) {
        Intent home = new Intent(context, Launcher.class);
        home.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP);
        home.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        context.startActivity(home);
    }


    public void backLauncherHome(Context context, String dialogTipString) {
        Intent home = new Intent(context, Launcher.class);
        home.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP);
        home.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        home.putExtra(DIALOG_KEY_VALUE, dialogTipString);
        context.startActivity(home);
    }

    public void backLauncherHome(Context context, int dialogType) {
        Intent home = new Intent(context, Launcher.class);
        home.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP);
        home.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        home.putExtra(DIALOG_KEY_VALUE, dialogType);
        context.startActivity(home);
    }

    public void jumpSetting(Context context) {
        Intent intent = new Intent(context, mSettingActivity);
        intent.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP);
        intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        context.startActivity(intent);
    }

    public static Context getContext() {
        return mContext;
    }
}
